/*
 * @DEC_COPYRIGHT@
 */
/*
 * HISTORY
 * $Log: win_utils.c,v $
 * Revision 1.1.2.2  1994/12/21  15:48:59  John_Dennis
 * 	new contributed code from SGI
 * 	[1994/12/21  15:42:35  John_Dennis]
 *
 * $EndLog$
 */
/*  
 *
 *  Windowing utilities for GLX.
 *
 *  The intent here is to expose the policies behind visual selection
 *  so you can tune them to fit your application.
 *
 *  get_overlay_visual() selects a transparent/opaque overlay visual 
 *  that matches the specified window attributes and get_main_visual() 
 *  selects a visual in the main planes, with the specified window
 *  attributes. get_colormap() uses the Xmu library to load a shared 
 *  colormap based on the visual; if a shared colormap is not available
 *  for the visual that was selected then a colormap is allocated.
 *
 *  You will find notes interspersed throughout the comments (search
 *  for NOTE:) indicating ways in which you may modify the
 *  visual selection criteria.
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include "win_utils.h"

static Atom wm_delete_window = None;

/*
 * Opens a GLX connection. If successful, this routine returns a
 * pointer to the display connection otherwise it returns NULL.
 */
Display *open_glx_connection( char *display)
{
    Display *dpy;
    int error, event;

    /* Open a connection to the X server */
    if ( !(dpy = XOpenDisplay( display)) ) {
       fprintf( stderr, "Cannot open display %s \n", XDisplayName( display));
       return NULL;
    }

    /* Make sure OpenGL is supported */
    if ( !glXQueryExtension( dpy, &error, &event) ) {
        fprintf( stderr, "No glx extension on %s \n", XDisplayName( display));
        return NULL;
    }

    return dpy;
}

/*  
 * Preference is given to the visual that supports the most colors
 * and also to pseudocolor over gray scale and true color over pseudocolor. 
 * Since there is a large number of colors available on 24 bit true/direct 
 * color devices, this algorithm heavily favors these visuals. (That is, 
 * the number of available colors is weighted much heavier than other factors 
 * such as depth buffer size.)
 */
static int score_visual_class ( XVisualInfo *visual)
{
    unsigned int mask;
    int score = 0;

    switch ( visual->class ) {
        case StaticGray:
        case GrayScale:
             /* favor pseudo color over gray scale */
             score = visual->colormap_size - 2;
             break;
        case StaticColor:
        case PseudoColor:
             /* favor true/direct over pseudo color */
             score = visual->colormap_size - 1;
             break;
        case TrueColor:
        case DirectColor:
            mask = visual->red_mask | 
		   visual->green_mask | 
		   visual->blue_mask;
	    if ( mask )
                for ( score = 1; mask > 0; mask >>= 1 ) if ( mask & 1 ) score <<= 1;
            break;
    }

    return( score);
}

/* 
 * Checks the visual attributes. If the specified attributes are not
 * supported then -1 is returned. If all the specified attributes are
 * supported then the visual is scored according to the depth of the
 * specified ancillary buffers, the depth of the color buffers, and
 * the type of visual. The score for the visual is returned.
 * (Note that alpha is treated differently than other visual attributes: 
 * if alpha is not specified then a visual with an alpha buffer will not 
 * be selected. See the comment in get_visual_and_colormap() for more details.)
 */
static int 
score_visual( Display *dpy, WinAttributes *attr, XVisualInfo *vis, WinAttributes *vis_attr)
{
    int val; 
    int score = 0;

    vis_attr->ancillary_buffers = 0;
   
    /* Color Index or RGBA? It must match the requested value */
    glXGetConfig( dpy, vis, GLX_RGBA, &vis_attr->color_mode);
    if ( attr->color_mode != WIN_DONTCARE &&
         attr->color_mode != vis_attr->color_mode ) return( -1);
   
    /* Stereo or mono visual? It must match the requested value */
    glXGetConfig( dpy, vis, GLX_STEREO, &vis_attr->stereo);
    if ( attr->stereo != WIN_DONTCARE &&
         attr->stereo != vis_attr->stereo ) return( -1);

    /* 
     * If double buffering is requested then visual must have a back buffer 
     * If single buffering is requested then it cannot have a back buffer   
     */
    glXGetConfig( dpy, vis, GLX_DOUBLEBUFFER, &val);
    vis_attr->ancillary_buffers |= (val) ? (WIN_DOUBLEBUFFER) : (WIN_SINGLEBUFFER);
    if ( (attr->ancillary_buffers & WIN_DOUBLEBUFFER) &&
         !(vis_attr->ancillary_buffers & WIN_DOUBLEBUFFER) ) return( -1);
    if ( (attr->ancillary_buffers & WIN_SINGLEBUFFER)  &&
         !(vis_attr->ancillary_buffers & WIN_SINGLEBUFFER) ) return( -1);

    /* 
     * If alpha requested then alpha size must be > 0. If alpha is not requested
     * then the size must be zero. 
     * NOTE: You may wish to check for a minimum alpha size. We assume
     * that size > 0 is sufficient if alpha is requested.
     */ 
    if ( vis_attr->color_mode == WIN_RGBA ) {
        glXGetConfig( dpy, vis, GLX_ALPHA_SIZE, &val);
        if ( val > 0 ) vis_attr->ancillary_buffers |= WIN_ALPHABUFFER;
        if ( (attr->ancillary_buffers & WIN_ALPHABUFFER) != 
             (vis_attr->ancillary_buffers & WIN_ALPHABUFFER) ) return( -1);
    }

    /* 
     * The number of requested aux buffers must be available 
     * NOTE: You may wish to adjust the score based on the number 
     * of available aux buffers.
     */
    glXGetConfig( dpy, vis, GLX_AUX_BUFFERS, &vis_attr->aux_buffers);
    if ( attr->aux_buffers &&
        vis_attr->aux_buffers < attr->aux_buffers ) return( -1);

    /* 
     * We just check that the specified ancillary buffers are supported;
     * that is the size just has to be > 0. If an ancillary buffer is 
     * supported then it's depth adds to the score. 
     * NOTE: You may want to make it possible to specify a minimum 
     * size for the ancillary buffers.
     */

    /* 
     * Accumulation buffers, if requested, are scored according the the 
     * size of the red, green, blue and alpha components.               
     */
    if ( vis_attr->color_mode == WIN_RGBA ) {
       glXGetConfig( dpy, vis, GLX_ACCUM_RED_SIZE, &val);
       if ( val > 0 )
           vis_attr->ancillary_buffers |= WIN_ACCUMBUFFER;
       if ( attr->ancillary_buffers & WIN_ACCUMBUFFER ) {
           if ( val <= 0 ) return( -1);
           score += val;
           glXGetConfig( dpy, vis, GLX_ACCUM_GREEN_SIZE, &val); score += val;
           glXGetConfig( dpy, vis, GLX_ACCUM_BLUE_SIZE, &val); score += val;
           glXGetConfig( dpy, vis, GLX_ACCUM_ALPHA_SIZE, &val); score += val;
       }
    }

    /* The depth buffer, if requested, is scored according to it's size */
    /* The visual is rejected if the size is zero                       */
    glXGetConfig( dpy, vis, GLX_DEPTH_SIZE, &val);
    if ( val > 0 ) 
        vis_attr->ancillary_buffers |= WIN_DEPTHBUFFER;
    if ( attr->ancillary_buffers & WIN_DEPTHBUFFER ) {
        if ( val <= 0 ) return( -1);
        score += val;
    }

    /* The stencil buffer is also scored according to it's size */
    /* The visual is rejected if the size is zero               */
    glXGetConfig( dpy, vis, GLX_STENCIL_SIZE, &val);
    if ( val > 0 ) 
        vis_attr->ancillary_buffers |= WIN_STENCILBUFFER;
    if ( attr->ancillary_buffers & WIN_STENCILBUFFER ) {
        if ( val <= 0 ) return( -1);
        score += val;
    }

    /* Score the visual class based on number of colors supported */
    score += score_visual_class( vis);

    return( score);
}


/* 
 * Finds the best main planes visual that supports all the requested attributes. 
 * (i.e., all requested ancillary buffers and aux buffers must be supported. Also 
 * if stereo, alpha planes, double buffering or single buffering is requested 
 * then the visual must support it. Note that, a double buffered visual will 
 * not be returned if single buffering is explicitly requested.) The caller 
 * should specify WIN_DONTCARE if they don't want care about the value of a
 * particular attribute.
 * 
 * If more than one visual satisfies the above criteria then preference
 * is given to the visual that supports the most colors and has the deepest
 * ancillary buffers. (But, as noted above, the number of available colors is 
 * considered more important than the depth of the requested ancillary
 * buffers.)
 */
static 
int get_main_visual( Display *dpy, WinAttributes *attr, XVisualInfo *vis_info, WinAttributes *vis_attr)
{
    int i, status, nvisuals, val, this_score, best_score;
    XVisualInfo *vis_list, *this_vis, *best_vis, sample_vis;
    WinAttributes this_vis_attr, best_vis_attr;

    /* Get list of visuals for this screen */
    sample_vis.screen = DefaultScreen( dpy);
    vis_list = XGetVisualInfo( dpy, VisualScreenMask, &sample_vis, &nvisuals);

    /* Loop through the visuals to find the one that has the best score */
    best_score = -1; best_vis = NULL;
    for ( i = 0; i < nvisuals; i++ ) {
      this_vis = &vis_list[i];
      this_score = 0;

      /* Visual must be supported by GLX */
      if ( glXGetConfig( dpy, this_vis, GLX_USE_GL, &val) ) continue;
      if ( !val ) continue;

      /* Visual must be in main planes which is level 0 */ 
      if ( glXGetConfig( dpy, this_vis, GLX_LEVEL, &val) ) continue;
      if ( val != 0 ) continue;
   
      /* 
       * Score the visual. If any of the requested buffers are not available 
       * the returned score will be negative.
       */
      this_score = score_visual( dpy, attr, this_vis, &this_vis_attr);
      if ( this_score < 0 ) continue;

      if ( this_score > best_score ) {
          best_score = this_score;
          best_vis = this_vis;
          best_vis_attr = this_vis_attr;
      }
    }

    status = GL_FALSE;
    if ( best_vis != NULL ) {
        *vis_info = *best_vis;
        *vis_attr = best_vis_attr;
        vis_attr->level = 0;
        vis_attr->transparent_type = WIN_OPAQUE;
        status = GL_TRUE;
    }

    if ( nvisuals > 0 ) XFree ( vis_list);
    return status;
}


/* 
 * Finds the best visual that lies in the overlay or underlay planes (depending
 * on the specified level) and supports all the requested attributes. (All 
 * requested ancillary buffers and aux buffers must be supported. Also 
 * if stereo, alpha planes, double buffering or single buffering is requested 
 * then the visual must support it. Note that, a double buffered visual will 
 * not be returned if single buffering is explicitly requested.) The caller
 * should specify WIN_DONTCARE if they don't want care about the value of a
 * particular attribute.
 *
 * Note that some systems may not support ancillary buffers for overlay and underlay 
 * visuals. Thus, specifying ancillary buffers may cause the visual selection
 * to fail.
 *
 * If more than one visual satisfies the above criteria then preference
 * is given to the visual that supports the most colors and has the deepest
 * ancillary buffers. (But, as noted above, the number of available colors is 
 * considered more important than the depth of the requested ancillary
 * buffers.)
 */
static int 
get_overlay_visual( Display *dpy, WinAttributes *attr, XVisualInfo *vis_info, WinAttributes *vis_attr)

{
   Atom overlay_property, actual_type;
   WinOverlayProperty *overlay_visuals;
   unsigned long item_count, num_items, bytes_unread;
   int noverlay_visuals, i, val, actual_format;
   XVisualInfo *this_vis, *best_vis, sample_vis;
   int this_score, best_score;
   WinAttributes this_vis_attr, best_vis_attr;

   /* If the atom doesn't exist then the overlay property isn't there */
   overlay_property = XInternAtom( dpy, "SERVER_OVERLAY_VISUALS", True);
   if ( overlay_property == None ) return GL_FALSE; 

   /* Fetch the property. We want a list of overlay visuals */
   bytes_unread = 0;
   item_count = sizeof( WinOverlayProperty)/4;
   do {
      item_count += (bytes_unread + 3)/4;
      XGetWindowProperty( dpy, RootWindow( dpy, DefaultScreen( dpy)),
         overlay_property, 0, item_count, False, overlay_property,
         &actual_type, &actual_format, &num_items, &bytes_unread,
         (unsigned char **) &overlay_visuals);
      if ( bytes_unread > 0 ) XFree( overlay_visuals);
   } while ( bytes_unread > 0 );

#define SKIP_VIS(this_vis) \
{ XFree( this_vis); continue; }

   /* Loop through all the overlay visuals to find the best one */
   best_score = -1;
   best_vis = NULL;
   noverlay_visuals = num_items/4;
   for ( i = 0; i < noverlay_visuals; i++ ) {
      sample_vis.visualid = overlay_visuals[i].visualid;
      this_vis = XGetVisualInfo( dpy, VisualIDMask, &sample_vis, &val);

      /* Visual must be supported by GLX */
      if ( glXGetConfig( dpy, this_vis, GLX_USE_GL, &val) ) SKIP_VIS( this_vis);
      if ( !val ) SKIP_VIS( this_vis);

      /* Level > 0 is overlay, level < 0 is underlay */ 
      if ( glXGetConfig( dpy, this_vis, GLX_LEVEL, &val) ) SKIP_VIS( this_vis);
      if ( attr->level != val ) SKIP_VIS( this_vis);

      /* If transparent/opaque specified then honor request */
      if ( overlay_visuals[i].transparent_type == None )
          this_vis_attr.transparent_type = WIN_OPAQUE;
      else
	  this_vis_attr.transparent_type = WIN_TRANSPARENT; 
      if ( attr->transparent_type != WIN_DONTCARE &&
           attr->transparent_type != this_vis_attr.transparent_type )
          SKIP_VIS( this_vis);

      /* 
       * Score the visual. If any of the requested buffers are not available 
       * the returned score will be negative.
       */
      this_score = score_visual( dpy, attr, this_vis, &this_vis_attr);
      if ( this_score < 0 ) SKIP_VIS( this_vis);

      if ( this_score > best_score ) {
          if ( best_vis != NULL ) XFree ( best_vis);
          best_vis = this_vis;
          best_vis_attr = this_vis_attr;
          best_score = this_score;
      }
    }

    if ( noverlay_visuals > 0 ) XFree( overlay_visuals);
    if ( best_vis != NULL ) {
        *vis_info = *best_vis;
        *vis_attr = best_vis_attr;
        vis_attr->level = attr->level;
        XFree( best_vis);
        return GL_TRUE;
    } else
        return GL_FALSE;
}

/*
 * Given a visual, choose an appropriate colormap. We use the Xmu 
 * utilities to make sure a standard colormap is loaded. 
 *
 * There are several types of standard colormaps:
 *           
 *   RGB_DEFAULT_MAP: The range of colors may be less than for 
 *       RGB_BEST_MAP. For example on an 8-plane system, a typical 
 *       allocation is 6 reds, 6 greens and 6 blues, yielding 216
 *       uniformly distributed colors. This leaves 40 elements of a 
 *       256 element colormap available for special purpose colors.
 *       The RGB_DEFAULT_MAP is usually available for all main plane 
 *       visuals. 
 *   RGB_BEST_MAP: This may only be available for one visual--usually a
 *       true color or direct color visual. It may support more distinct
 *       colors than RGB_DEFAULT_MAP. 
 *   RGB_RED_MAP, RGB_GREEN_MAP, RGB_BLUE_MAP: These colormaps have all-red,
 *       all-green or all-blue components. This is useful if you want to
 *       make color separated images
 *   RGB_GRAY_MAP: A gray scale colormap.
 *
 * The code below always tries to load the default map. Why? The gray
 * red, green and blue maps, aren't generally desirable for 3D graphics.
 * And the best map--if it is supported on this visual--is probably the 
 * same as the default map. 
 * 
 * If a standard colormap isn't available for this visual, then we create
 * a colormap (Most likely, this will only happen in the case of overlay and 
 * underlay visuals)
 */
static 
Colormap get_colormap( Display *dpy, XVisualInfo *vis_info)
{
   Atom wm_property;
   XStandardColormap *std_cmaps;
   int num_cmaps, i;
   Colormap cmap;

   /* 
    * NOTE: You may want to check for RGB_BEST_MAP first and then, if it isn't
    * available, try to get the the default map
    */

   wm_property = XA_RGB_DEFAULT_MAP;

   if ( XmuLookupStandardColormap( dpy, vis_info->screen, vis_info->visualid,
           vis_info->depth, wm_property, False/*replace*/, True/*retain*/) ) {
      if ( XGetRGBColormaps( dpy, RootWindow( dpy, vis_info->screen),
           &std_cmaps, &num_cmaps, wm_property) ) {
         for (i = 0; i < num_cmaps; i++) {
            if ( std_cmaps[i].visualid == vis_info->visualid ) {
               cmap = std_cmaps[i].colormap;
               XFree( std_cmaps);
	       return cmap;
            }
         }
         if ( num_cmaps > 0 ) XFree( std_cmaps);
      }
   }

   return( XCreateColormap( dpy, RootWindow( dpy, vis_info->screen), 
           vis_info->visual, AllocNone));
}


/* 
 * Finds the best visual in the overlay/underlay/main planes (depending
 * on the level attribute) and a colormap to match.
 */
int get_visual_and_colormap(Display *dpy, int mask, WinAttributes *attr, XVisualInfo 
                            *vis_info, Colormap *cmap)
{
    int status;
    WinAttributes all_attr;
  
    all_attr.color_mode = WIN_DONTCARE;
    all_attr.stereo = WIN_DONTCARE;
    all_attr.ancillary_buffers = 0;
    all_attr.aux_buffers = 0;
    all_attr.level = 0;
    all_attr.transparent_type = WIN_DONTCARE;

    if ( mask & WIN_COLOR_MODE_ATTR )
        all_attr.color_mode = attr->color_mode;
    if ( mask & WIN_STEREO_ATTR )
        all_attr.stereo = attr->stereo;
    if ( mask & WIN_ANCILLARY_BUFFERS_ATTR )
        all_attr.ancillary_buffers = attr->ancillary_buffers;
    if ( mask & WIN_AUX_BUFFERS_ATTR )
        all_attr.aux_buffers = attr->aux_buffers;
    if ( mask & WIN_LEVEL_ATTR )
        all_attr.level = attr->level;
    if ( mask & WIN_TRANSPARENT_TYPE_ATTR )
        all_attr.transparent_type = attr->transparent_type;
    
    /* If color index is requested then disable requests for alpha and accum */
    /* NOTE: You may want to flag this as an error                           */
    if ( all_attr.color_mode == WIN_INDEX )
        all_attr.ancillary_buffers &= ~(WIN_ALPHABUFFER | WIN_ACCUMBUFFER);

    if ( all_attr.level != 0 )
        status = get_overlay_visual( dpy, &all_attr, vis_info, attr);
    else
        status = get_main_visual( dpy, &all_attr, vis_info, attr);

    /* 
     * Visual selection can fail if all the matching visuals have alpha buffers
     * and an alpha buffer was not specified. Thus, if an alpha buffer was not
     * specified originally, we try again this time explicitly asking for an 
     * alpha buffer. 
     *
     * Why special case alpha buffers? Many graphics adaptors export GLX visuals 
     * with alpha buffers and GLX visuals without alpha buffers. On these adaptors 
     * it is may be the case that that choosing a visual with an alpha buffer will 
     * slow down rendering (since the alpha values must be written into the color 
     * buffer). Thus, we will only select a visual with an alpha buffer when alpha 
     * is explicitly requested or when there are no visuals without it.
     */
    if ( !status ) {
       if ( (attr->color_mode == WIN_INDEX) || (all_attr.ancillary_buffers & WIN_ALPHABUFFER) ) 
          return GL_FALSE;
       all_attr.ancillary_buffers |= WIN_ALPHABUFFER;
       if ( all_attr.level != 0 )
          status = get_overlay_visual( dpy, &all_attr, vis_info, attr);
       else
          status = get_main_visual( dpy, &all_attr, vis_info, attr);
       if ( !status ) return GL_FALSE;
    }

    *cmap = get_colormap( dpy, vis_info);
    return GL_TRUE;
}


/* 
 * Creates a window with the specified visual and colormap and parent. The 
 * specified position, size and title are used to set window property hints.
 */
Window create_window(Display *dpy, Window parent_win, char *title_string, 
     XVisualInfo *vis_info, Colormap cmap, int x, int y, unsigned int width, 
     unsigned int height)
{
    Window win;
    XSizeHints win_hint; /* window manager hints */
    XSetWindowAttributes win_attr;

    win_hint.x = x; win_hint.y = y;
    win_hint.width = width; win_hint.height = height;
    win_hint.flags = PPosition | PSize;

    /* Create the window using the specified visual and colormap */
    /* Have to specify border pixel if visual is not the default */
    win_attr.colormap = cmap;
    win_attr.border_pixel = 0;
    win = XCreateWindow( dpy, parent_win,
        win_hint.x, win_hint.y, win_hint.width, win_hint.height,
        0 /*border width*/, vis_info->depth, InputOutput, vis_info->visual,
        CWColormap | CWBorderPixel, &win_attr);

   /* Now set the window title and size hints */
   XSetStandardProperties( dpy, win, title_string, title_string, None, NULL,
        0, &win_hint);

   /* Allow clean termination from window manager */
   if ( wm_delete_window == None )
       wm_delete_window = XInternAtom( dpy, "WM_DELETE_WINDOW", False);
   if ( wm_delete_window != None )
       XSetWMProtocols( dpy, win, &wm_delete_window, 1);

   return win;

}

int decode_client_message( XEvent *event)
{
   if (event->xclient.data.l[0] == wm_delete_window)
       return WIN_CLIENT_SHUTDOWN;
   return 0;
}

