/*
 * @DEC_COPYRIGHT@
 */
/*
 * HISTORY
 * $Log: polytest.c,v $
 * Revision 1.1.2.5  1995/05/05  20:15:18  John_Dennis
 * 	remove old tesselation code
 * 	[1995/05/05  20:14:30  John_Dennis]
 *
 * Revision 1.1.2.4  1995/05/05  19:20:33  John_Dennis
 * 	bring GLU up to V1.2
 * 	[1995/05/05  19:14:46  John_Dennis]
 * 
 * Revision 1.1.2.3  1994/12/28  15:52:27  John_Dennis
 * 	fix typo bug where y coord was being assigned to z value.
 * 	[1994/12/28  15:51:57  John_Dennis]
 * 
 * Revision 1.1.2.2  1994/11/15  15:59:38  John_Dennis
 * 	Original version
 * 	[1994/11/15  15:58:53  John_Dennis]
 * 
 * $EndLog$
 */

#if 0
                             -- DOCUMENTATION --

Program to test GLU polygon tesselation. Written by John Dennis.

The program reads a polygon definition file which must be specfied with
-f filename on the command line. The file format is:


Blank lines are ignored.

From the character '#' to the end of the line is a comment.

A line with the string New Contour introduces a new contour. The New
Contour command is optional for the first contour.

A line with 2 or three numbers is a vertex, if only two numbers are specfied
then the missing number is taken to be z=0.

Notes:

Need gluTessBeginContour and gluTessEndContour, this is not backward
compatible. Missing calls generate errors.

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <float.h>
#include <GL/glx.h>
#include <GL/glu.h>

static int attributeList[] = { GLX_RGBA,
			    GLX_RED_SIZE, 1,
			    GLX_GREEN_SIZE, 1,
			    GLX_BLUE_SIZE, 1,
			    GLX_DOUBLEBUFFER,
			    GLX_DEPTH_SIZE, 16,
			    None };

/* command line args */
char *progName = NULL;
int debug = 0;
int edges = 0;

/* bounding box */
float bb_left   =  FLT_MAX;
float bb_right  = -FLT_MAX;
float bb_bottom =  FLT_MAX;
float bb_top    = -FLT_MAX;
float bb_near   =  FLT_MAX;
float bb_far    = -FLT_MAX;

/*
** A contour.
*/
typedef struct Vert
{
  float x;
  float y;
  float z;
  struct Vert *next;
} Vert;
/*
** A polygon (multiple contours).
*/
typedef struct Polygon {
  Vert *contour;
  int nVertices;
  struct Polygon *next;
} Polygon;

/* triangulation object */
static GLUtesselator *tessObj = NULL;

static void initgfx(void);
static void setsize(int width, int height);
static void drawit(void);
static void checkerrors(void);
static Vert *addVertex(Vert **next, GLfloat x, GLfloat y, GLfloat z);
static Polygon *newContour(void);
static void dumpPolygon(Polygon *thePoly);
static void freePolygon(Polygon *thePoly);

void cbBegin(GLenum type);
void cbEdgeFlag(GLboolean flag);
void cbVertex(void *data);
void cbEnd(void);
void cbError(GLenum error);
void cbCombine( GLdouble coords[3], Vert *data[4],
                GLfloat weight[4], Vert **dataOut );

Display* dpy;
Window win;
GLXContext cx;
int width, height;
Polygon *thePoly = NULL;
Vert *combineVerts = NULL;

void
Usage()
{
  fprintf(stderr, "%s: -f filename [args]\n\n", progName);
  
  fprintf(stderr, "args:\n");
  fprintf(stderr, "-help        print usage instructions\n");
  fprintf(stderr, "-f filename  read polygon definition from filename\n");
  fprintf(stderr, "-edges       turn on edges\n");
  fprintf(stderr, "-debug n     set debug level to n, higher values more verbose\n");

  fprintf(stderr, "\nkeys:\n\n");
  fprintf(stderr, "<ESC>        exit program:\n");
}


int main(int argc, char **argv) {
  XVisualInfo* vi;
  Colormap cmap;
  XSetWindowAttributes swa;
  XEvent event;
  XButtonEvent* button;
  XKeyEvent* key;
  XMotionEvent* motion;
  XExposeEvent* expose;
  XConfigureEvent* config;
  char s[100];
  int i;
  char *filename = NULL;
    

  progName = argv[0];
  for(i=1; i < argc; i++) {
    if (!strcmp(argv[i], "-f")) {
      if (++i >= argc) {
        fprintf(stderr, "no file specfied for -f\n");
        Usage();
        exit(1);
      }
      filename = argv[i];
    }
    else if (!strcmp(argv[i], "-edges")) {
      edges = 1;
    }
    else if (!strcmp(argv[i], "-debug")) {
      if (++i >= argc) {
        fprintf(stderr, "no debug level specfied for -debug\n");
        Usage();
        exit(1);
      }
      debug = atoi(argv[i]);
    }
    else if (!strcmp(argv[i], "-help")) {
        Usage();
        exit(1);
    }
    else {
      fprintf(stderr, "unknown arg: %s\n", argv[i]);
      Usage();
      exit(1);
    }
  }
  if (!filename) {
    fprintf(stderr, "no filename specfied\n");
    Usage();
    exit(1);
  }
  ReadFile(filename);
  if (debug >= 2) dumpPolygon(thePoly);

  /* get a connection */
  dpy = XOpenDisplay(0);

  /* get an appropriate visual */
  vi = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList);

  /* create a GLX context */
  cx = glXCreateContext(dpy, vi, 0, GL_TRUE);

  /* create a color map */
  cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen),
                         vi->visual, AllocNone);

  /* create a window */
  swa.colormap = cmap;
  swa.border_pixel = 0;
  swa.event_mask = StructureNotifyMask;
  width = 400;
  height = 400;
  win = XCreateWindow(dpy, RootWindow(dpy, vi->screen), 0, 0, width, height,
                      0, vi->depth, InputOutput, vi->visual,
                      CWBorderPixel|CWColormap|CWEventMask, &swa);
  XMapWindow(dpy, win);

  /* add program name to window banner */
  XStoreName(dpy, win, argv[0]);

  /* prepare for event loop */
  button = (XButtonEvent*)(&event);
  key = (XKeyEvent*)(&event);
  motion = (XMotionEvent*)(&event);
  expose = (XExposeEvent*)(&event);
  config = (XConfigureEvent*)(&event);
  XSelectInput(dpy, win, KeyPressMask |
               ButtonPressMask |
               ButtonReleaseMask |
               Button1MotionMask |
               Button2MotionMask |
               ExposureMask |
               StructureNotifyMask);

  /* loop handling events */
  while (1) {
    XNextEvent(dpy, &event);
    switch (event.type) {
    case KeyPress:
      i = XLookupString(key, s, 100, NULL, NULL);
      if (i) {
        switch ((int)(s[0])) {
        case 033:
          gluDeleteTess(tessObj);
          freePolygon(thePoly);
          thePoly = NULL;
          XFree(vi);
          glXMakeCurrent(dpy, None, NULL);
          glXDestroyContext(dpy, cx);
          XCloseDisplay(dpy);
          exit(0);
          break;
        default:
          s[i] = 0;
          /* printf("%s, %d\n", s, i); */
          break;
        }
      }
      break;
    case ButtonPress:
      /* printf("%d button pressed\n", button->button); */
      break;
    case ButtonRelease:
      /* printf("%d button released\n", button->button); */
      break;
    case MotionNotify:
      /* printf("%3d,%3d tracking\n", motion->x, motion->y); */
      break;
    case Expose:
      /* printf("expose %d\n", expose->count); */
      /* printf("  x=%d, y=%d\n", expose->x, expose->y); */
      /* printf("  width=%d, height=%d\n", */
      /*	expose->width, expose->height); */
      if (expose->count == 0) {
        initgfx();
        drawit();
      }
      break;
    case ConfigureNotify:
      width = config->width;
      height = config->height;
      initgfx();
      setsize(width, height);
      drawit();
      break;
    }
  }
}

static void initgfx(void) {
  static done = 0;
  float width, height, depth, left, right, bottom, top, near, far;
    
  if (!done) {
    done = 1;
    glXMakeCurrent(dpy, win, cx);
    glClearColor(0,0,0,1);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    if ((width = bb_right - bb_left) <= 0.0) width = 1.0;
    if ((height = bb_top - bb_bottom) <= 0.0) height = 1.0;
    if ((depth = bb_far - bb_near) <= 0.0) depth = 1.0;
    
    left   =   bb_left   - width  * 0.2;
    right  =   bb_right  + width  * 0.2;
    bottom =   bb_bottom - height * 0.2;
    top    =   bb_top    + height * 0.2;
    near   =   bb_near   - depth  * 0.2;
    far    =   bb_far    + depth  * 0.2;

    if (debug >= 1)
      printf("glOrtho: left=%f right=%f bottom=%f top=%f near=%f far=%f\n",
             left, right, bottom, top, near, far);

    glOrtho(left, right, bottom, top, near, far);
    
    tessObj = gluNewTess();

    if (edges) {
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      gluTessCallback(tessObj, GLU_EDGE_FLAG, cbEdgeFlag);
    }
    
    gluTessCallback(tessObj, GLU_BEGIN, cbBegin);
    gluTessCallback(tessObj, GLU_VERTEX, cbVertex);
    gluTessCallback(tessObj, GLU_END, cbEnd);
    gluTessCallback(tessObj, GLU_ERROR, cbError);
    gluTessCallback(tessObj, GLU_TESS_COMBINE, &cbCombine);
  }
}

static void setsize(int width, int height) {
  glViewport(0, 0, width, height);
}

static void drawit(void) {
  int i;
  
  glClear(GL_COLOR_BUFFER_BIT);

  Triangulate();

  glXSwapBuffers(dpy, win);
  checkerrors();
}

static void checkerrors(void) {
  int i;
  int goterror;

  goterror = 0;
  while (i = glGetError()) {
    goterror = 1;
    fprintf(stderr, "ERROR: 0x%x %s\n", i, gluErrorString(i));
  }
  if (goterror)
    exit(1);
}

static Polygon *newContour(void)
{
  Polygon *newPoly;
  Polygon **polyTail;

  newPoly = (Polygon *) malloc(sizeof(Polygon));

  polyTail = &thePoly;
  while (*polyTail) {
    polyTail = &((*polyTail)->next);
  }

  *polyTail = newPoly;
  newPoly->next = NULL;
  newPoly->contour = NULL;
  newPoly->nVertices = 0;
  return(newPoly);
}

static Vert *addVertex(Vert **next, GLfloat x, GLfloat y, GLfloat z)
{
  Vert *newVert;

  newVert = (Vert *) malloc(sizeof(Vert));

  newVert->next = *next;
  *next = newVert;
  newVert->x = x;
  newVert->y = y;
  newVert->z = z;

  return newVert;
}

int
ReadFile(char *filename)
{
  char *p, buf[1024];
  FILE *fp;
  float x,y,z;
  int nRead;
  int lineNum;
  int nVertices = 0;
  int nContours = 0;
  Polygon *poly;
  Vert *vert, **pVert;

  if (!(fp = fopen(filename, "r"))) {
    fprintf(stderr, "can't open file %s for read\n%s\n",
            filename, sys_errlist[errno]);
    exit(1);
  }
  

  poly = newContour();
  pVert = &poly->contour;
  for(lineNum=1; fgets(buf, 1024, fp); lineNum++) {
    for (p=buf; *p && isspace(*p); p++); /* skip whitespace */
    if (!*p) continue;          /* blank line */
    if (*p == '#') continue;    /* comment */
    if (!strncmp(p, "New Contour", strlen("New Contour"))) {
      if (poly->nVertices) {
        poly = newContour();
        pVert = &poly->contour;
        nContours++;
        if (debug >= 2) printf("\nNew Contour %d\n", nContours);
      }
      continue;
    }
    
    nRead = sscanf(buf, "%f %f %f", &x, &y, &z);
    switch(nRead) {
    case 2:
      z = 0.0;
      break;
    case 3:
      break;
    default:
      fprintf(stderr, "read error line %d, unknown input, expecting coordinate\n%s\n",
              lineNum, buf);
      exit(1);
    }
    if (x < bb_left) bb_left = x;
    if (x > bb_right) bb_right = x;
    if (y < bb_bottom) bb_bottom = y;
    if (y > bb_top) bb_top = y;
    if (z < bb_near) bb_near = z;
    if (z > bb_far) bb_far = z;

    vert = addVertex(pVert, x, y, z);
    pVert = &vert->next;
    poly->nVertices++;
    nVertices++;
    
    if (debug >= 2) printf("line %3d: %f %f %f\n", lineNum, x, y, z);
    
  }
  if (debug >=1)
    printf("bounding box: left=%f right=%f bottom=%f top=%f near=%f far=%f\n",
           bb_left, bb_right, bb_bottom, bb_top, bb_near, bb_far);

  
  fclose(fp);
  return(0);
}

static void dumpPolygon(Polygon *thePoly)
{
    
  Polygon *poly;
  Vert *vert;
  int nContours = 0;
  int nVertices = 0;

  for (poly=thePoly; poly; poly=poly->next, nContours++) {
    printf("Contour %d nVertices=%d\n", nContours, poly->nVertices);
    for (vert=poly->contour, nVertices=0; vert; vert=vert->next, nVertices++) {
      printf("    %3d %6.4f %6.4f %6.4f\n",
             nVertices, vert->x, vert->y, vert->z);
    }
    printf("\n");
  }
}

static void freePolygon(Polygon *thePoly)
{
    
  Polygon *poly, *nextPoly;
  Vert *vert, *nextVert;

  for (poly=thePoly; poly; poly=nextPoly) {
    nextPoly = poly->next;
    for (vert=poly->contour; vert; vert=nextVert) {
      nextVert = vert->next;
      free(vert);
    }
    free(poly);
  }

  for (vert=combineVerts; vert; vert=nextVert) {
    nextVert = vert->next;
    free(vert);
  }
}

int
Triangulate()
{
  Polygon *poly;
  Vert *vert;
  int nContours = 0;
  int nVertices = 0;
  GLdouble vertex[3];

  
  if (debug >= 2) printf("\nTesselating\n");
  
  gluTessBeginPolygon(tessObj, NULL);
  for (poly=thePoly; poly; poly=poly->next, nContours++) {
    if (debug >= 2) printf("Contour %d\n", nContours);
    gluTessBeginContour(tessObj);  
    for (vert=poly->contour, nVertices=0; vert; vert=vert->next, nVertices++) {
      vertex[0] = vert->x;
      vertex[1] = vert->y;
      vertex[2] = vert->z;
      if (debug >= 2)
        printf("tess vertex %3d: %f %f %f\n",
               nVertices, vertex[0], vertex[1], vertex[2]);

      gluTessVertex(tessObj, vertex, vert);
    }
    gluTessEndContour(tessObj);
  }

  gluTessEndPolygon(tessObj);
}

void
cbBegin (GLenum type)
{
  printf ("cbBegin:     ");
  switch (type) {
  case GL_TRIANGLES:
    printf ("GL_TRIANGLES\n");
    glBegin (GL_TRIANGLES);
    break;
  case GL_TRIANGLE_STRIP:
    printf ("GL_TRIANGLE_STRIP\n");
    glBegin (GL_TRIANGLE_STRIP);
    break;
  case GL_TRIANGLE_FAN:
    printf ("GL_TRIANGLE_FAN\n");
    glBegin (GL_TRIANGLE_FAN);
    break;
  default:
    printf("NOT A TRIANGULAR FORM\n");
    break; 
  }
}

void
cbEdgeFlag (GLboolean flag)
{
  printf ("cbEdgeFlag:  ");
  glEdgeFlag(flag);
  switch (flag) {
  case GL_TRUE:
    printf("GL_TRUE\n");
    break;
  case GL_FALSE:
    printf("GL_FALSE\n");
    break;
  default:
    printf("garbage flag = %d\n", flag);
    break;
  }
}


void
cbVertex (void *data)
{
  Vert *pt_data;
  
  pt_data = (Vert *) data;
  printf ("cbVertex:    ");
  printf ("%f, %f, %f\n", pt_data->x, pt_data->y, pt_data->z);
  glVertex3f(pt_data->x, pt_data->y, pt_data->z);
}

void
cbEnd (void)
{
  printf ("cbEnd\n");
  glEnd();
}

void
cbError (GLenum error)
{
  printf ("cbError\n  ");
  printf ("        Error = %d %s\n", error, gluErrorString(error));
}

void cbCombine( GLdouble coords[3], Vert *data[4],
                     GLfloat weight[4], Vert **dataOut )
{
  Vert *vertex;

  printf ("cbCombine    ");
  printf ("%f, %f, %f\n",  coords[0],coords[1],coords[2]);
  vertex = addVertex(&combineVerts, coords[0],coords[1],coords[2]);
  *dataOut = vertex;
}

